////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// mc12106load.h -                                                            //
//                                                                            //
//      Load and exchange library function declaration                        //
//      (for one-processor board MC121.06)                                    //
//                                                                            //
// Copyright (c) 2020 RC Module                                               //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#ifndef __MC12106LOAD__H__
#define __MC12106LOAD__H__

#include <linux/types.h>

#define PLOAD_VERSION "1.1"
#define PLOAD_MAJOR_VERSION 1
#define PLOAD_MINOR_VERSION 1

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

	// 32 bit unsigned NM memory element type
	typedef __u32 PL_Word;

	// 32 bit unsigned NM address type
	typedef __u32 PL_Addr;

	// Board identification variants
	enum PL_identification_variant
	{
		PL_ARBITRARY_INSTANCE              = 0,
		PL_IDENTIFICATION_BY_COM_PORT_NAME = 1,
		PL_IDENTIFICATION_BY_BOARD_ID      = 2
	};

	// Definition of board identification variant
	typedef struct _PL_board_identification
	{
		int identification_variant;
		const char *com_port_name;
		__u32 board_id;
	} PL_board_identification;

	// Board descriptor
	typedef struct PL_Board PL_Board;

	// Processor on board descriptor
	typedef struct PL_Access PL_Access;

	// Library functions return values (all library functions return result code)
	enum RetValue
	{
		PL_OK              = 0, // OK
		PL_ERROR           = 1, // Error
		PL_TIMEOUT         = 2, // Timeout of wait for operation
		PL_FILE            = 3, // Cannot find file for load
		PL_BADADDRESS      = 4, // Bad address ranges
		PL_NOT_IMPLEMENTED = -1 // Not implemented
	};

	// Flag of the user's program work end
	const PL_Word PROGRAM_FINISHED = 0x2;

	//------------------//
	// Common functions //
	//------------------//

	// Return version of library in variable pointed by 'version'
	int PL_GetVersion(unsigned int *version);

	// Return number of detected boards in variable pointed by 'count'
	int PL_GetBoardCount(unsigned int *count);

	// Create descriptor for board
	// Return descriptor in variable pointed by 'board'
	int PL_GetBoardDescEx(PL_board_identification *board_identification, PL_Board **board);

	// Close board descriptor and free descriptor memory
	int PL_CloseBoardDesc(PL_Board *board);

	//-----------------//
	// Board functions //
	//-----------------//

	// Get board ID
	int PL_GetBoardId(PL_Board *board, unsigned long *board_id);

	// Get version of firmware of board
	int PL_GetFirmwareVersion(PL_Board *board, unsigned int *version_major, unsigned int *version_minor);

	// Send RESET signal to board and all processors on board
	int PL_ResetBoard(PL_Board *board);

	// Load init code to board
	// Call this function after reset board and before loading user program to processors.
	int PL_LoadInitCode(PL_Board *board);

	// Create descriptor for processor number 'procNo' on board
	// Return processor descriptor in variable pointed by 'access'
	// Processor number is 0
	int PL_GetAccess(PL_Board *board, unsigned int procNo, PL_Access **access);

	// Close processor descriptor
	int PL_CloseAccess(PL_Access *access);

	//---------------------//
	// Processor functions //
	//---------------------//

	// Load user program on processor from memory and start execution
	int PL_LoadProgram(PL_Access *access, const void *addrProgram, unsigned int sizeProgram);

	// Load user program on processor from file and start execution
	int PL_LoadProgramFile(PL_Access *access, const char *filename);

	// Read array from shared memory
	// block   - Pointer to destination buffer in PC memory
	// len     - Size of array in 32-bit words
	// address - Address of source array in NM memory
	int PL_ReadMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, PL_Word len);

	// Write array in shared memory
	// block   - Pointer to source array in PC memory
	// len     - Size of array in 32-bit words
	// address - Address of destination array in NM memory
	int PL_WriteMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, PL_Word len);

	// Barrier syncronization with program on board processor
	// value       - value sent to processor
	// returnValue - value received from processor
	int PL_Sync(PL_Access *access, int value, int *returnValue);

	// Barrier syncronization with program on board processor
	// Additionally perform two values exchange
	// value       - value sent to processor
	// outAddress  - address sent to processor
	// outLen      - size sent to processor
	// returnValue - value received from processor
	// inAddress   - address received from processor
	// inLen       - size received from processor
	// Values return only if pointers are not NULL
	int PL_SyncArray(
		PL_Access *access,  // Processor descriptor

		int value,          // Value sent to processor
		PL_Addr outAddress, // Address sent to processor
		PL_Word outLen,     // Size sent to processor

		int *returnValue,   // Value received from processor
		PL_Addr *inAddress, // Address received from processor
		PL_Word *inLen);    // Size received from processor

	// Set timeout in milliseconds for waiting function
	// Used in all PL_SyncXXX functions
	int PL_SetTimeout(__u32 timeout);

	// Return status bits in variable 'status'
	int PL_GetStatus(PL_Access *access, PL_Word *status);

	// Return user program return value in variable 'result'
	int PL_GetResult(PL_Access *access, PL_Word *result);

	//------------------------------------------//
	// Special functions only in MC121.06 board //
	//------------------------------------------//

	// Getting count of boards in STM mode and their IDs
	int PL_GetStmCount(unsigned int *stm_count, __u32 *boards_ids, unsigned int boards_ids_size);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif // __MC12106LOAD__H__
