// Load and exchange library for MC121.06

#include <stdio.h>
#include <string.h>
#include <sstream>
#include <limits.h>
#include <libudev.h>
#include <pthread.h>
#include <sys/time.h>
#include <assert.h>

#include "libelf.h"

#include "libload_def.h"
#include "libload_int.h"

// Timeout for SyncXXX functions (mS)
// (0 - Infinite waiting)
static __u32 globalTimeout = 0;

// For synchronization access to LibLoad
pthread_mutex_t LL_mutex = PTHREAD_MUTEX_INITIALIZER;

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

static int ReadMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, PL_Word len);
static int WriteMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, PL_Word len);
static int GetStatus(PL_Access *access, PL_Word *status);

//------------------//
// Common functions //
//------------------//

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetVersion -                                                            //
//      Getting library version                                               //
//                                                                            //
// Return values:                                                             //
//      Always PL_OK                                                          //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetVersion(unsigned int *version)
{
	PL_DBG("Begin.");

	if (version)
		*version = PLOAD_MAJOR_VERSION;

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetBoardCount -                                                         //
//      Getting count of boards                                               //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetBoardCount(unsigned int *count)
{
	unsigned int stm_cnt = 0, nm_cnt = 0;

	PL_DBG("Begin.");

	int cbl = Capture_common_boards_lock();
	if (cbl == -1)
	{
		PL_DBG("ERROR: Failed capture common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (Enum_boards(STM_PRODUCT_ID, NULL, &stm_cnt, NULL, NULL))
	{
		PL_DBG("ERROR: Failed get count of STM!");
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (Enum_boards(NM_PRODUCT_ID, NULL, &nm_cnt, NULL, NULL))
	{
		PL_DBG("ERROR: Failed get count of NM!");
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (Release_common_boards_lock(cbl))
	{
		PL_DBG("ERROR: Failed release common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (count)
		*count = stm_cnt + nm_cnt;

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetBoardDescEx -                                                        //
//      Getting descriptor for access to board                                //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetBoardDescEx(
	PL_board_identification *board_identification,
	PL_Board **board)
{
	PL_DBG("Begin.");

	if (!board_identification)
	{
		PL_DBG("ERROR: Invalid 'board_identification' (NULL)!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("Identification variant: %d", board_identification->identification_variant);

	switch (board_identification->identification_variant)
	{
	case PL_ARBITRARY_INSTANCE:

		PL_DBG("Identification variant: Arbitrary instance.");

		break;

	case PL_IDENTIFICATION_BY_COM_PORT_NAME:

		PL_DBG("Identification variant: Identification by COM-port.");

		if (!board_identification->com_port_name)
		{
			PL_DBG("ERROR: Invalid 'board_identification->com_port_name' (NULL)!");
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		PL_DBG("COM-port: %s", board_identification->com_port_name);

		break;

	case PL_IDENTIFICATION_BY_BOARD_ID:

		PL_DBG("Identification variant: Identification by board ID.");

		PL_DBG("Board ID: 0x%8.8X = %u", board_identification->board_id, board_identification->board_id);

		break;

	default:
		PL_DBG("ERROR: Invalid board identification variant value!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	} // switch

	int cbl = Capture_common_boards_lock();
	if (cbl == -1)
	{
		PL_DBG("ERROR: Failed capture common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("Search STM COM-port...");

	char stm_com_port_name[PATH_MAX + 1];
	__u32 stm_id;
	bool stm_com_port_found = false;

	if (Enum_boards(STM_PRODUCT_ID, board_identification, NULL, &stm_com_port_found, stm_com_port_name))
	{
		PL_DBG("ERROR: Failed enumerate STM boards!");
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	char nm_com_port_name[PATH_MAX + 1];
	bool nm_com_port_found = false;

	if (stm_com_port_found)
	{
		PL_DBG("STM COM-port was found.");

		if (board_identification->identification_variant == PL_IDENTIFICATION_BY_BOARD_ID)
			stm_id = board_identification->board_id;
		else
		{
			PL_DBG("Getting STM ID...");

			if (Send_command_to_stm(stm_com_port_name, STM_GET_ID, &stm_id))
			{
				PL_DBG("ERROR: Failed get STM ID!");
				Release_common_boards_lock(cbl);
				PL_DBG("End (With ERROR!); Return: PL_ERROR!");
				return PL_ERROR;
			}
		}

		PL_DBG("stm_id = 0x%8.8X = %u", stm_id, stm_id);

		PL_DBG("Switching to NM...");

		if (Send_command_to_stm(stm_com_port_name, STM_SWITCH_TO_NM, NULL))
		{
			PL_DBG("ERROR: Failed switch STM to NM!");
			Release_common_boards_lock(cbl);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		PL_DBG("Waiting for NM COM-port to appear...");

		for (unsigned int i = 0; i < 100; i++)
		{
			PL_DBG("i = %u", i);

			usleep(100000);

			PL_board_identification bi = {PL_IDENTIFICATION_BY_BOARD_ID, NULL, stm_id};

			if (Enum_boards(NM_PRODUCT_ID, &bi, NULL, &nm_com_port_found, nm_com_port_name))
			{
				PL_DBG("ERROR: Failed enumerate NM boards!");
				Release_common_boards_lock(cbl);
				PL_DBG("End (With ERROR!); Return: PL_ERROR!");
				return PL_ERROR;
			}

			if (nm_com_port_found)
				break;
		} // for i
	} // if (stm_com_port_found)
	else
	{
		PL_DBG("STM COM-port was not found.");
		PL_DBG("Search NM COM-port...");

		if (Enum_boards(NM_PRODUCT_ID, board_identification, NULL, &nm_com_port_found, nm_com_port_name))
		{
			PL_DBG("ERROR: Failed enumerate NM boards!");
			Release_common_boards_lock(cbl);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}
	} // if not (stm_com_port_found)

	if (!nm_com_port_found)
	{
		PL_DBG("ERROR: Compliant NM COM-port not found!");
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("nm_com_port_name = '%s'", nm_com_port_name);

	PL_Board *newBoard;

	try
	{
		newBoard = new PL_Board(nm_com_port_name);
	}
	catch (...)
	{
		PL_DBG("ERROR: Failed create board descriptor!");
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board)
	{
		if (board_identification->identification_variant == PL_IDENTIFICATION_BY_BOARD_ID)
			newBoard->Board_ID = board_identification->board_id;
		else if (stm_com_port_found)
			newBoard->Board_ID = stm_id;
		else
			if (newBoard->Get_board_id())
			{
				PL_DBG("ERROR: Failed get board ID!");
				delete newBoard;
				Release_common_boards_lock(cbl);
				PL_DBG("End (With ERROR!); Return: PL_ERROR!");
				return PL_ERROR;
			}

		PL_DBG("Board ID: 0x%8.8X = %u", newBoard->Board_ID, newBoard->Board_ID);

		*board = newBoard;
	}
	else
	{
		// Result not need, deleting descriptor
		delete newBoard;
	}

	if (Release_common_boards_lock(cbl))
	{
		PL_DBG("ERROR: Failed release common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_CloseBoardDesc -                                                        //
//      Closing descriptor for access to board                                //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_CloseBoardDesc(PL_Board *board)
{
	PL_DBG("Begin.");

	LOCK_LL

	if (Check_board(board))
	{
		PL_DBG("ERROR: Invalid 'board'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board->Count_opened_PL_Access)
	{
		PL_DBG("ERROR: Board has %u opened processor descriptor(s)!", board->Count_opened_PL_Access);
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board->Count_working_functions)
	{
		PL_DBG("ERROR: Board has %u working functions!", board->Count_working_functions);
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	board->Magic_code = 0x0;

	UNLOCK_LL

	int cbl = Capture_common_boards_lock();
	if (cbl == -1)
	{
		PL_DBG("ERROR: Failed capture common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board->Switch_to_stm_if_no_more_used_and_close_com_port())
	{
		PL_DBG("WARNING: Failed switch to STM!");
	}

	delete board;

	if (Release_common_boards_lock(cbl))
	{
		PL_DBG("ERROR: Failed release common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//-----------------------------------//
// Functions for work with one board //
//-----------------------------------//

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetBoardId -                                                            //
//      Getting board ID                                                      //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetBoardId(
	PL_Board *board,
	unsigned long *board_id)
{
	PL_DBG("Begin.");

	if (!board_id)
	{
		PL_DBG("ERROR: Invalid 'board_id' (NULL)!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	LOCK_LL

	if (Check_board(board))
	{
		PL_DBG("ERROR: Invalid 'board'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	*board_id = board->Board_ID;

	UNLOCK_LL

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetFirmwareVersion -                                                    //
//      Getting version of firmware of board                                  //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetFirmwareVersion(
	PL_Board *board,
	unsigned int *version_major,
	unsigned int *version_minor)
{
	PL_DBG("Begin.");

	if ((!version_major) || (!version_minor))
	{
		PL_DBG("ERROR: Invalid 'version_major' (%p) and / or 'version_minor' (%p)!", version_major, version_minor);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (Check_and_capture_board(board))
	{
		PL_DBG("ERROR: Failed check and capture board!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board->Capture_port())
	{
		PL_DBG("ERROR: Failed capture COM-port!");
		Release_board(board);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board->Send_command(READ_MEM, 0, FIRMWARE_VERSION_BASE, 2))
	{
		PL_DBG("ERROR: Failed send command (for version reading)!");
		board->Release_port();
		Release_board(board);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_Word version[2];

	if (Read_from_com_port(board->hCom_port, (char *)version, 8))
	{
		PL_DBG("ERROR: Failed read version from COM-port!");
		board->Release_port();
		Release_board(board);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (board->Release_port())
	{
		PL_DBG("ERROR: Failed release COM-port!");
		Release_board(board);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	Release_board(board);

	*version_major = version[MAJOR_OFFSET];
	*version_minor = version[MINOR_OFFSET];

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_ResetBoard -                                                            //
//      Reset board                                                           //
//                                                                            //
// Notes:                                                                     //
//      Not implemented for this board.                                       //
//      Always return PL_NOT_IMPLEMENTED.                                     //
//                                                                            //
// Return values:                                                             //
//      Always PL_NOT_IMPLEMENTED                                             //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_ResetBoard(PL_Board *board)
{
	PL_DBG("Start.");

	PL_DBG("End; Return: PL_NOT_IMPLEMENTED.");
	return PL_NOT_IMPLEMENTED;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_LoadInitCode -                                                          //
//      Load to board init code                                               //
//                                                                            //
// Notes:                                                                     //
//      Does nothing.                                                         //
//      Always return PL_OK.                                                  //
//                                                                            //
// Return values:                                                             //
//      Always PL_OK                                                          //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_LoadInitCode(PL_Board *board)
{
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetAccess -                                                             //
//      Getting descriptor for access to processor                            //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetAccess(
	PL_Board *board,
	unsigned int procNo,
	PL_Access **access)
{
	PL_DBG("Begin; procNo = %u.", procNo);

	if (procNo >= CORES_COUNT)
	{
		PL_DBG("ERROR: Invalid 'procNo'!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	LOCK_LL

	if (Check_board(board))
	{
		PL_DBG("ERROR: Invalid 'board'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_Access *newAccess;

	try
	{
		newAccess = new PL_Access(procNo, board);
	}
	catch (...)
	{
		PL_DBG("ERROR: Failed create processor descriptor!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access)
	{
		*access = newAccess;
	}
	else
	{
		// Result not need, deleting descriptor
		delete newAccess;
	}

	UNLOCK_LL

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_CloseAccess -                                                           //
//      Closing descriptor for access to processor                            //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_CloseAccess(PL_Access *access)
{
	PL_DBG("Begin.");

	LOCK_LL

	if (Check_access(access))
	{
		PL_DBG("ERROR: Invalid 'access'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->Count_working_functions)
	{
		PL_DBG("ERROR: Access has %u working functions!", access->Count_working_functions);
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->Sync_in_progress)
	{
		PL_DBG("ERROR: Sync in progress (Sync_in_progress = %d)!", access->Sync_in_progress);
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	access->Magic_code = 0x0;

	UNLOCK_LL

	delete access;

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//------------------------------------------------//
// Functions for work with one processor on board //
//------------------------------------------------//

// Prototype for internal function
static int LoadProgramFile(
	PL_Access *access,
	bool load_from_file,
	const char *filename,
	const void *addrProgram,
	unsigned int sizeProgram,
	int testKernelRange);

//----------------------------------------------------------------------------//
//                                                                            //
// PL_LoadProgram -                                                           //
//      Load and run user program from memory                                 //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//      PL_FILE - User program file not found or not be an ELF-file           //
//      PL_BADADDRESS - Incorrect deployment segments in memory               //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_LoadProgram(
	PL_Access *access,
	const void *addrProgram,
	unsigned int sizeProgram)
{
	return LoadProgramFile(access, false, NULL, addrProgram, sizeProgram, 1);
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_LoadProgramFile -                                                       //
//      Load and run user program                                             //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//      PL_FILE - User program file not found or not be an ELF-file           //
//      PL_BADADDRESS - Incorrect deployment segments in memory               //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_LoadProgramFile(
	PL_Access *access,
	const char *filename)
{
	return LoadProgramFile(access, true, filename, NULL, 0, 1);
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_LoadProgramFile_Int -                                                   //
//      Load and run debugger resident                                        //
//      (Hidden function.)                                                    //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//      PL_FILE - User program file not found or not be an ELF-file           //
//      PL_BADADDRESS - Incorrect deployment segments in memory               //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_LoadProgramFile_Int(
	PL_Access *access,
	const char *filename)
{
	return LoadProgramFile(access, true, filename, NULL, 0, 0);
}

//----------------------------------------------------------------------------//
//                                                                            //
// LoadProgramFile -                                                          //
//      Load and run user program                                             //
//                                                                            //
// Notes:                                                                     //
//      Depending on testKernelRange execute address range checking of        //
//      loading program:                                                      //
//      0 - no checking;                                                      //
//      1 - standard checking;                                                //
//      (Internal function.)                                                  //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//      PL_FILE - User program file not found or not be an ELF-file           //
//      PL_BADADDRESS - Incorrect deployment segments in memory               //
//                                                                            //
//----------------------------------------------------------------------------//

int LoadProgramFile(
	PL_Access *access,
	bool load_from_file,
	const char *filename,
	const void *addrProgram,
	unsigned int sizeProgram,
	int testKernelRange)
{
	int ret;

	PL_DBG("Begin; load_from_file = %d; testKernelRange = %d.", load_from_file, testKernelRange);

	if (load_from_file)
	{
		PL_DBG("Loading from file.");

		if (!filename)
		{
			PL_DBG("ERROR: Invalid 'filename' (NULL)!");
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		PL_DBG("filename = '%s'", filename);
	}
	else
	{
		PL_DBG("Loading from memory.");
		PL_DBG("addrProgram = %p; sizeProgram = %u", addrProgram, sizeProgram);

		if (!addrProgram || !sizeProgram)
		{
			PL_DBG("ERROR: Invalid 'addrProgram' and / or 'sizeProgram'!");
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}
	}

	if (Check_and_capture_access(access))
	{
		PL_DBG("ERROR: Failed check and capture access!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->board->Capture_port())
	{
		PL_DBG("ERROR: Failed capture COM-port!");
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_Word status;

	// Getting loader status
	if (GetStatus(access, &status) != PL_OK)
	{
		PL_DBG("ERROR: Failed get status!");
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// Checking loader status
	if (status & PROGRAM_STARTED)
	{
		PL_DBG("ERROR: Program already started (status = 0x%8.8X)!", status);
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	ELF *ElfFile = NULL;

	// Reading abs
	try
	{
		if (load_from_file)
			ElfFile = new ELF(filename, ELF_C_READ);
		else
			ElfFile = new ELF((char *)addrProgram, sizeProgram, ELF_C_READ);
	}
	catch (ELF_ERROR& ErrorCodeElf)
	{
		// File not found

		const char *ErrorCodeString;
		std::ostringstream os;
		os << ErrorCodeElf;
		std::string s;
		s = os.str();
		ErrorCodeString = s.c_str();

		PL_DBG("ERROR: ELF ERROR: %s!", ErrorCodeString);
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_FILE!");
		return PL_FILE;
	}
	catch (...)
	{
		PL_DBG("ERROR: ERROR detected!");
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	try
	{
		// If this is not individual ELF-file
		if (ElfFile->Kind() != ELF_K_ELF)
		{
			throw ELF_ERROR("Wrong kind of file");
		}

		// Reading file header
		Elf32_Ehdr *Ehdr = ElfFile->GetEhdr();

		// If wrong machine type
		if (!((Ehdr->e_machine == EM_NM640X) &&
			((Ehdr->e_flags == EM_NM640X_NMC4) || (Ehdr->e_flags == EM_NM640X_NMC4I) || (Ehdr->e_flags == EM_NM640X_NMC4F))))
		{
			PL_DBG("e_machine = %u; e_flags = %u", Ehdr->e_machine, Ehdr->e_flags);
			throw ELF_ERROR("Wrong machine type");
		}

		// If not absolute executable file
		if (Ehdr->e_type != ET_EXEC)
		{
			throw ELF_ERROR("Bad file type");
		}

		// If there is no program segments
		if (!Ehdr->e_phnum)
		{
			throw ELF_ERROR("Program have not program segments");
		}

		// Reading table of program segments
		Elf32_Phdr *Phdr = ElfFile->GetPhdr();

		// Loading segments
		for (__u32 i = 0; i < Ehdr->e_phnum; i++) // Cycle by segments
		{
			// If not loading segment
			if (Phdr[i].p_type != PT_LOAD)
				continue;

			if (!(Phdr[i].p_memsz >> 2))
				continue;

			PL_DBG("Loading segment %u:", i);
			PL_DBG("Address:            0x%8.8X", Phdr[i].p_paddr);
			PL_DBG("Length (in file):   0x%8.8X", Phdr[i].p_filesz >> 2);
			PL_DBG("Length (in memory): 0x%8.8X", Phdr[i].p_memsz >> 2);

			if (Check_mem_ranges(access->processorNo, Phdr[i].p_paddr, Phdr[i].p_memsz >> 2, testKernelRange))
			{
				throw ELF_ERROR("Bad address ranges in program");
			}

#ifdef FILL_WITH_ZERO
			__u32 blockSize = Phdr[i].p_memsz;
			PL_DBG("Segments fill with zero at end (blockSize = 0x%8.8X).", blockSize);
#else
			__u32 blockSize = Phdr[i].p_filesz;
			PL_DBG("Segments not fill with zero at end (blockSize = 0x%8.8X).", blockSize);
#endif

			if (!blockSize)
				continue;

			// Buffer for reading program segment
			char *Buffer = new char[blockSize];

			try
			{
				// Loading segment
#ifdef FILL_WITH_ZERO
				ElfFile->LoadProgramSegment(&Phdr[i], Buffer);
#else
				ElfFile->LoadProgramSegment(&Phdr[i], Buffer, 1);
#endif

				blockSize = blockSize >> 2;

				// Convert to view in memory
				ELF::ToM(
					ELF32_T_WORD,
					(unsigned char *)Buffer,
					(unsigned char *)Buffer,
					ELFDATA2LSB,
					blockSize);

				ret = WriteMemBlock(access, (__u32 *)Buffer, Phdr[i].p_paddr, blockSize);
				if (ret != PL_OK)
				{
					PL_DBG("ERROR: Failed write to memory (ret = %d)!", ret);
					throw 1;
				}
			}
			catch (...)
			{
				PL_DBG("ERROR: ERROR detected!");
				delete[] Buffer;
				throw;
			}
			delete[] Buffer;
		} // From cycle by segments

		// Start address of user program
		PL_Addr StartAddr = Ehdr->e_entry;

		// Setting entry point of program and start program

		if (access->board->Send_command(RUN_PROGRAM, access->processorNo, StartAddr, testKernelRange))
		{
			PL_DBG("ERROR: Failed start program!");
			throw 1;
		}

		delete ElfFile;
	}
	catch (ELF_ERROR& ErrorCodeElf)
	{
		// Loading error
		delete ElfFile;

		const char *ErrorCodeString;
		std::ostringstream os;
		os << ErrorCodeElf;
		std::string s;
		s = os.str();
		ErrorCodeString = s.c_str();

		PL_DBG("ERROR: ELF ERROR: %s!", ErrorCodeString);
		if (!strcmp(ErrorCodeString, "Bad address ranges in program"))
		{
			access->board->Release_port();
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
			return PL_BADADDRESS;
		}
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}
	catch (...)
	{
		PL_DBG("ERROR: ERROR detected!");
		delete ElfFile;
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->board->Release_port())
	{
		PL_DBG("ERROR: Failed release COM-port!");
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	Release_access(access);

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_ReadMemBlock -                                                          //
//      Reading memory block                                                  //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_ReadMemBlock(
	PL_Access *access,
	PL_Word *block,
	PL_Addr address,
	PL_Word len)
{
	PL_DBG("Begin; address = 0x%8.8X; len = 0x%8.8X.", address, len);

	if (!block)
	{
		PL_DBG("ERROR: Invalid 'block' (NULL)!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (Check_and_capture_access(access))
	{
		PL_DBG("ERROR: Failed check and capture access!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// If len = 0 then only checking need, no transfer will be confirm.
	// In this case for correct checking address parameter len must be 1.

	if (Check_mem_ranges(access->processorNo, address, len ? len : 1, 0))
	{
		PL_DBG("ERROR: Invalid 'address' (0x%8.8X) and / or 'len' (0x%8.8X)!", address, len);
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
		return PL_BADADDRESS;
	}

	int ret = PL_OK;

	if (len)
	{
		if (access->board->Capture_port())
		{
			PL_DBG("ERROR: Failed capture COM-port!");
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		ret = ReadMemBlock(access, block, address, len);

		if (access->board->Release_port())
		{
			PL_DBG("ERROR: Failed release COM-port!");
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}
	}

	Release_access(access);

	PL_DBG("End; Return: %d.", ret);
	return ret;
}

static int ReadMemBlock(
	PL_Access *access,
	PL_Word *block,
	PL_Addr address,
	PL_Word len)
{
	PL_DBG("Begin; address = 0x%8.8X; len = 0x%8.8X.", address, len);

	assert(access);
	assert(block);
	assert(len);

	if (access->Read(block, address, len))
	{
		PL_DBG("ERROR: Failed read memory!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_WriteMemBlock -                                                         //
//      Writing memory block                                                  //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_WriteMemBlock(
	PL_Access *access,
	PL_Word *block,
	PL_Addr address,
	PL_Word len)
{
	PL_DBG("Begin; address = 0x%8.8X; len = 0x%8.8X.", address, len);

	if (!block)
	{
		PL_DBG("ERROR: Invalid 'block' (NULL)!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (Check_and_capture_access(access))
	{
		PL_DBG("ERROR: Failed check and capture access!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// If len = 0 then only checking need, no transfer will be confirm.
	// In this case for correct checking address parameter len must be 1.

	if (Check_mem_ranges(access->processorNo, address, len ? len : 1, 1))
	{
		PL_DBG("ERROR: Invalid 'address' (0x%8.8X) and / or 'len' (0x%8.8X)!", address, len);
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
		return PL_BADADDRESS;
	}

	int ret = PL_OK;

	if (len)
	{
		if (access->board->Capture_port())
		{
			PL_DBG("ERROR: Failed capture COM-port!");
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		ret = WriteMemBlock(access, block, address, len);

		if (access->board->Release_port())
		{
			PL_DBG("ERROR: Failed release COM-port!");
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}
	}

	Release_access(access);

	PL_DBG("End; Return: %d.", ret);
	return ret;
}

static int WriteMemBlock(
	PL_Access *access,
	PL_Word *block,
	PL_Addr address,
	PL_Word len)
{
	PL_DBG("Begin; address = 0x%8.8X; len = 0x%8.8X.", address, len);

	assert(access);
	assert(block);
	assert(len);

	if (access->Write(block, address, len))
	{
		PL_DBG("ERROR: Failed write memory!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_Sync -                                                                  //
//      Barrier synchronization with NM, scalar                               //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_Sync(
	PL_Access *access,
	int value,
	int *returnValue)
{
	return PL_SyncArray(
		access,
		value,
		0,
		0,
		returnValue,
		NULL,
		NULL);
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_SyncArray -                                                             //
//      Barrier synchronization with NM, array                                //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_SyncArray(
	PL_Access *access,

	int value,
	PL_Addr outAddress,
	PL_Word outLen,

	int *returnValue,
	PL_Addr *inAddress,
	PL_Word *inLen)
{
	PL_Word Buf[4];

	PL_DBG("Begin; value = %d; outAddress = 0x%8.8X; outLen = 0x%8.8X.", value, outAddress, outLen);

	if (Check_and_capture_access_for_sync(access))
	{
		PL_DBG("ERROR: Failed check and capture access!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->board->Capture_port())
	{
		PL_DBG("ERROR: Failed capture COM-port!");
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// Writting values (Output)

	if (access->Write1(SYNC_FROM_PC_BASE + SYNC_VALUE_OFFSET, value))
	{
		PL_DBG("ERROR: Failed write to memory!");
		access->board->Release_port();
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->Write1(SYNC_FROM_PC_BASE + SYNC_ADDR_OFFSET, outAddress))
	{
		PL_DBG("ERROR: Failed write to memory!");
		access->board->Release_port();
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->Write1(SYNC_FROM_PC_BASE + SYNC_LEN_OFFSET, outLen))
	{
		PL_DBG("ERROR: Failed write to memory!");
		access->board->Release_port();
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// Setting synchro bit (Output)
	if (access->Write1(SYNC_FROM_PC_BASE + SYNC_FLAG_OFFSET, NC_SYNC_FLAG))
	{
		PL_DBG("ERROR: Failed write to memory!");
		access->board->Release_port();
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// Waiting of setting synchro bit (Input)

	struct timeval tv;
	gettimeofday(&tv, NULL);
	unsigned int start_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
	unsigned int current_time;

	while (1)
	{
		if (access->Read(Buf, SYNC_TO_PC_BASE, 4))
		{
			PL_DBG("ERROR: Failed read from memory!");
			access->board->Release_port();
			Release_access_for_sync(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		if (Buf[SYNC_FLAG_OFFSET] == NC_SYNC_FLAG)
			break;

		if (globalTimeout)
		{
			gettimeofday(&tv, NULL);
			current_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

			if ((current_time - start_time) > globalTimeout)
			{
				access->board->Release_port();
				Release_access_for_sync(access);
				PL_DBG("End; Return: PL_TIMEOUT.");
				return PL_TIMEOUT;
			}
		}

		usleep(1000);
	}

	// Reading values (Input)

	if (access->Read(Buf, SYNC_TO_PC_BASE, 4))
	{
		PL_DBG("ERROR: Failed read from memory!");
		access->board->Release_port();
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (returnValue)
		*returnValue = Buf[SYNC_VALUE_OFFSET];

	if (inAddress)
		*inAddress = Buf[SYNC_ADDR_OFFSET];

	if (inLen)
		*inLen = Buf[SYNC_LEN_OFFSET];

	// Clearing synchro bit (Input)
	if (access->Write1(SYNC_TO_PC_BASE + SYNC_FLAG_OFFSET, 0))
	{
		PL_DBG("ERROR: Failed write to memory!");
		access->board->Release_port();
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	// Waiting of clearing synchro bit (Output)

	gettimeofday(&tv, NULL);
	start_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

	while (1)
	{
		if (access->Read(Buf, SYNC_FROM_PC_BASE, 4))
		{
			PL_DBG("ERROR: Failed read from memory!");
			access->board->Release_port();
			Release_access_for_sync(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		if (!Buf[SYNC_FLAG_OFFSET])
			break;

		if (globalTimeout)
		{
			gettimeofday(&tv, NULL);
			current_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

			if ((current_time - start_time) > globalTimeout)
			{
				access->board->Release_port();
				Release_access_for_sync(access);
				PL_DBG("End; Return: PL_TIMEOUT.");
				return PL_TIMEOUT;
			}
		}

		usleep(1000);
	}

	if (access->board->Release_port())
	{
		PL_DBG("ERROR: Failed release COM-port!");
		Release_access_for_sync(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	Release_access_for_sync(access);

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_SetTimeout -                                                            //
//      Setting timeout                                                       //
//                                                                            //
// Notes:                                                                     //
//      Specify in mS. 0 - infinite waiting.                                  //
//      Used in all SyncXXX functions.                                        //
//                                                                            //
// Return value:                                                              //
//      Always PL_OK                                                          //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_SetTimeout(__u32 timeout)
{
	PL_DBG("Begin; timeout = %u.", timeout);

	LOCK_LL

	globalTimeout = timeout;

	UNLOCK_LL

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetStatus -                                                             //
//      Getting loader status                                                 //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetStatus(
	PL_Access *access,
	PL_Word *status)
{
	PL_DBG("Begin.");

	if (Check_and_capture_access(access))
	{
		PL_DBG("ERROR: Failed check and capture access!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	int ret = PL_OK;

	if (status)
	{
		if (access->board->Capture_port())
		{
			PL_DBG("ERROR: Failed capture COM-port!");
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		ret = GetStatus(access, status);

		if (access->board->Release_port())
		{
			PL_DBG("ERROR: Failed release COM-port!");
			Release_access(access);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}
	}

	Release_access(access);

	PL_DBG("End; Return: %d.", ret);
	return ret;
}

static int GetStatus(
	PL_Access *access,
	PL_Word *status)
{
	PL_DBG("Begin.");

	assert(access);
	assert(status);

	PL_Word tmp[4];

	if (access->Read(tmp, LOADER_ADDR_BASE, 4))
	{
		PL_DBG("ERROR: Failed get status!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	*status = tmp[STATUS_WORD_OFFSET];

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetResult -                                                             //
//      Getting return value of user program                                  //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetResult(
	PL_Access *access,
	PL_Word *result)
{
	PL_DBG("Begin.");

	if (Check_and_capture_access(access))
	{
		PL_DBG("ERROR: Failed check and capture access!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->board->Capture_port())
	{
		PL_DBG("ERROR: Failed capture COM-port!");
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_Word tmp[4];

	if (access->Read(tmp, LOADER_ADDR_BASE, 4))
	{
		PL_DBG("ERROR: Failed get status and return value!");
		access->board->Release_port();
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (access->board->Release_port())
	{
		PL_DBG("ERROR: Failed release COM-port!");
		Release_access(access);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	Release_access(access);

	if (!(tmp[STATUS_WORD_OFFSET] & PROGRAM_STOPPED))
	{
		PL_DBG("ERROR: User program not complete (Status: 0x%8.8X)!", tmp[STATUS_WORD_OFFSET]);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	if (result)
		*result = tmp[RETURN_VALUE_OFFSET];

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

//------------------------------------------//
// Special functions only in MC121.06 board //
//------------------------------------------//

//----------------------------------------------------------------------------//
//                                                                            //
// PL_GetStmCount -                                                           //
//      Getting count of boards in STM mode and their IDs                     //
//                                                                            //
// Return values:                                                             //
//      PL_OK - Success                                                       //
//      PL_ERROR - Error                                                      //
//                                                                            //
//----------------------------------------------------------------------------//

int PL_GetStmCount(
	unsigned int *stm_count,
	__u32 *boards_ids,
	unsigned int boards_ids_size)
{
	struct udev *udev_context;
	struct udev_enumerate *udev_enum;
	struct udev_list_entry *enum_list, *enum_list_first;
	struct udev_device *udev_dev;
	int ret;

	PL_DBG("Begin; boards_ids_size = %u.", boards_ids_size);

	int cbl = Capture_common_boards_lock();
	if (cbl == -1)
	{
		PL_DBG("ERROR: Failed capture common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	udev_context = udev_new();
	if (!udev_context)
	{
		PL_DBG("ERROR: Failed to allocate udev context!");
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	udev_enum = udev_enumerate_new(udev_context);
	if (!udev_enum)
	{
		PL_DBG("ERROR: Failed to allocate udev enumeration!");
		udev_unref(udev_context);
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	ret = udev_enumerate_add_match_subsystem(udev_enum, "tty");
	if (ret < 0)
	{
		PL_DBG("ERROR: Failed to add match (ret = %d)!", ret);
		udev_enumerate_unref(udev_enum);
		udev_unref(udev_context);
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	ret = udev_enumerate_add_match_property(udev_enum, "ID_VENDOR_ID", VENDOR_ID);
	if (ret < 0)
	{
		PL_DBG("ERROR: Failed to add match (ret = %d)!", ret);
		udev_enumerate_unref(udev_enum);
		udev_unref(udev_context);
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	ret = udev_enumerate_scan_devices(udev_enum);
	if (ret < 0)
	{
		PL_DBG("ERROR: Failed to scan devices (ret = %d)!", ret);
		udev_enumerate_unref(udev_enum);
		udev_unref(udev_context);
		Release_common_boards_lock(cbl);
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	enum_list_first = udev_enumerate_get_list_entry(udev_enum);

	PL_DBG("enum_list_first = %p", enum_list_first);

	unsigned int nIndex = 0;
	unsigned int cnt = 0;

	udev_list_entry_foreach(enum_list, enum_list_first)
	{
		PL_DBG("nIndex = %u; cnt = %u", nIndex, cnt);

		nIndex++;

		udev_dev = udev_device_new_from_syspath(udev_context, udev_list_entry_get_name(enum_list));
		if (!udev_dev)
		{
			PL_DBG("ERROR %d (%s): Failed to allocate udev device!", errno, strerror(errno));
			udev_enumerate_unref(udev_enum);
			udev_unref(udev_context);
			Release_common_boards_lock(cbl);
			PL_DBG("End (With ERROR!); Return: PL_ERROR!");
			return PL_ERROR;
		}

		const char *dev_name = udev_device_get_property_value(udev_dev, "DEVNAME");
		const char *vid = udev_device_get_property_value(udev_dev, "ID_VENDOR_ID");
		const char *pid = udev_device_get_property_value(udev_dev, "ID_MODEL_ID");

		PL_DBG("Device name: '%s'", dev_name);
		PL_DBG("Vendor ID: %s", vid);
		PL_DBG("Product ID: %s", pid);

		if (strcmp(vid, VENDOR_ID)) // For any case
		{
			PL_DBG("ERROR: Incorrect vendor ID!");
			udev_device_unref(udev_dev);
			assert(false);
			continue;
		}

		if (strcmp(pid, STM_PRODUCT_ID))
		{
			udev_device_unref(udev_dev);
			continue;
		}

		PL_DBG("Board found.");

		if (cnt < boards_ids_size)
		{
			PL_DBG("Getting STM ID...");

			if (Send_command_to_stm(dev_name, STM_GET_ID, &boards_ids[cnt]))
			{
				PL_DBG("ERROR: Failed get STM ID!");
				udev_device_unref(udev_dev);
				udev_enumerate_unref(udev_enum);
				udev_unref(udev_context);
				Release_common_boards_lock(cbl);
				PL_DBG("End (With ERROR!); Return: PL_ERROR!");
				return PL_ERROR;
			}

			PL_DBG("Board ID: 0x%8.8X = %u", boards_ids[cnt], boards_ids[cnt]);
		}

		cnt++;

		udev_device_unref(udev_dev);
	} // udev_list_entry_foreach

	udev_enumerate_unref(udev_enum);
	udev_unref(udev_context);

	if (Release_common_boards_lock(cbl))
	{
		PL_DBG("ERROR: Failed release common boards lock!");
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	PL_DBG("cnt = %u", cnt);

	if (stm_count)
		*stm_count = cnt;

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus
