// Load and exchange library for MC121.06
// Internal functions

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/file.h>
#include <limits.h>
#include <libudev.h>
#include <pthread.h>
#include <assert.h>
#include <errno.h>

#include "libload_def.h"
#include "libload_int.h"

extern pthread_mutex_t LL_mutex;

int Capture_common_boards_lock()
{
	PL_DBG("Begin.");

	int cbl = open(COMMON_BOARD_LOCK_FILE, O_CREAT, S_IRUSR | S_IRGRP | S_IROTH);
	if (cbl == -1)
	{
		PL_DBG("ERROR %d (%s): Failed create or open common boards lock file!", errno, strerror(errno));
		PL_DBG("End (With ERROR!); Return: -1!");
		return -1;
	}

	PL_DBG("Capturing common boards lock...");

	int ret = flock(cbl, LOCK_EX);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed capture common boards lock (ret = %d)!", errno, strerror(errno), ret);
		close(cbl);
		PL_DBG("End (With ERROR!); Return: -1!");
		return -1;
	}

	PL_DBG("End; Return: %d.", cbl);
	return cbl;
}

int Release_common_boards_lock(int cbl)
{
	PL_DBG("Begin.");

	int ret = close(cbl);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed close common boards lock file (ret = %d)!", errno, strerror(errno), ret);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

static int Enum_boards_(const char *product_id, PL_board_identification *board_identification, unsigned int *count, bool *com_port_found, char *com_port_name)
{
	struct udev *udev_context;
	struct udev_enumerate *udev_enum;
	struct udev_list_entry *enum_list, *enum_list_first;
	struct udev_device *udev_dev;
	int ret;

	PL_DBG("Begin; product_id = '%s'; board_identification = %p.", product_id, board_identification);

	udev_context = udev_new();
	if (!udev_context)
	{
		PL_DBG("ERROR: Failed to allocate udev context!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	udev_enum = udev_enumerate_new(udev_context);
	if (!udev_enum)
	{
		PL_DBG("ERROR: Failed to allocate udev enumeration!");
		udev_unref(udev_context);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	ret = udev_enumerate_add_match_subsystem(udev_enum, "tty");
	if (ret < 0)
	{
		PL_DBG("ERROR: Failed to add match (ret = %d)!", ret);
		udev_enumerate_unref(udev_enum);
		udev_unref(udev_context);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	ret = udev_enumerate_add_match_property(udev_enum, "ID_VENDOR_ID", VENDOR_ID);
	if (ret < 0)
	{
		PL_DBG("ERROR: Failed to add match (ret = %d)!", ret);
		udev_enumerate_unref(udev_enum);
		udev_unref(udev_context);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	ret = udev_enumerate_scan_devices(udev_enum);
	if (ret < 0)
	{
		PL_DBG("ERROR: Failed to scan devices (ret = %d)!", ret);
		udev_enumerate_unref(udev_enum);
		udev_unref(udev_context);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	enum_list_first = udev_enumerate_get_list_entry(udev_enum);

	PL_DBG("enum_list_first = %p", enum_list_first);

	unsigned int nIndex = 0;
	unsigned int cnt = 0;

	udev_list_entry_foreach(enum_list, enum_list_first)
	{
		PL_DBG("nIndex = %u", nIndex);

		nIndex++;

		udev_dev = udev_device_new_from_syspath(udev_context, udev_list_entry_get_name(enum_list));

		int ret_err = errno;

		PL_DBG("udev_dev = %p; ret_err = %d", udev_dev, ret_err);

		if (!udev_dev && (ret_err == ENODEV || ret_err == ENOENT))
		{
			PL_DBG("The device has disappeared. Need to wait for a new device to appear.");
			udev_enumerate_unref(udev_enum);
			udev_unref(udev_context);
			PL_DBG("End; Return: 2.");
			return 2;
		}

		if (!udev_dev)
		{
			PL_DBG("ERROR %d (%s): Failed to allocate udev device!", ret_err, strerror(ret_err));
			udev_enumerate_unref(udev_enum);
			udev_unref(udev_context);
			PL_DBG("End (With ERROR!); Return: 1!");
			return 1;
		}

		const char *dev_name = udev_device_get_property_value(udev_dev, "DEVNAME");
		const char *vid = udev_device_get_property_value(udev_dev, "ID_VENDOR_ID");
		const char *pid = udev_device_get_property_value(udev_dev, "ID_MODEL_ID");

		PL_DBG("Device name: '%s'", dev_name);
		PL_DBG("Vendor ID: %s", vid);
		PL_DBG("Product ID: %s", pid);

		if (strcmp(vid, VENDOR_ID)) // For any case
		{
			PL_DBG("ERROR: Incorrect vendor ID!");
			udev_device_unref(udev_dev);
			assert(false);
			continue;
		}

		if (strcmp(pid, product_id))
		{
			udev_device_unref(udev_dev);
			continue;
		}

		PL_DBG("Board found.");

		cnt++;

		if (board_identification)
		{
			bool appr_brd = false;

			switch (board_identification->identification_variant)
			{
			case PL_ARBITRARY_INSTANCE:

				PL_DBG("Case: Arbitrary instance.");

				appr_brd = true;

				break;

			case PL_IDENTIFICATION_BY_COM_PORT_NAME:

				PL_DBG("Case: Identification by COM-port (%s).", board_identification->com_port_name);

				if (!strcmp(dev_name, board_identification->com_port_name))
				{
					PL_DBG("The found board corresponds to.");

					appr_brd = true;
				}

				break;

			case PL_IDENTIFICATION_BY_BOARD_ID:

				PL_DBG("Case: Identification by board ID (0x%8.8X = %u).", board_identification->board_id, board_identification->board_id);

				__u32 board_id;

				if (!strcmp(product_id, STM_PRODUCT_ID))
				{
					PL_DBG("Getting STM ID...");

					if (Send_command_to_stm(dev_name, STM_GET_ID, &board_id))
					{
						PL_DBG("WARNING: Failed get STM ID!");
						break;
					}
				}
				else
				{
					PL_DBG("Getting NM ID...");

					assert(!strcmp(product_id, NM_PRODUCT_ID));

					if (Get_nm_id(dev_name, &board_id))
					{
						PL_DBG("WARNING: Failed get NM ID!");
						break;
					}
				}

				PL_DBG("board_id = 0x%8.8X = %u", board_id, board_id);

				if (board_id == board_identification->board_id)
				{
					PL_DBG("The found board corresponds to.");

					appr_brd = true;
				}

				break;

			default:
				PL_DBG("ERROR: Incorrect identification variant (%d)!", board_identification->identification_variant);
				udev_device_unref(udev_dev);
				udev_enumerate_unref(udev_enum);
				udev_unref(udev_context);
				assert(false);
				PL_DBG("End (With ERROR!); Return: 1!");
				return 1;
			} // switch

			if (appr_brd)
			{
				if (com_port_name)
				{
					strncpy(com_port_name, dev_name, PATH_MAX + 1);
				}

				*com_port_found = true;

				PL_DBG("COM-port (%s) found.", dev_name);

				udev_device_unref(udev_dev);
				udev_enumerate_unref(udev_enum);
				udev_unref(udev_context);

				PL_DBG("End; Return: 0.");
				return 0;
			}
		} // if (board_identification)

		udev_device_unref(udev_dev);
	} // udev_list_entry_foreach

	udev_enumerate_unref(udev_enum);
	udev_unref(udev_context);

	if (board_identification)
	{
		*com_port_found = false;

		PL_DBG("COM-port not found.");
	}

	PL_DBG("cnt = %u", cnt);

	if (count)
		*count = cnt;

	PL_DBG("End; Return: 0.");
	return 0;
}

int Enum_boards(const char *product_id, PL_board_identification *board_identification, unsigned int *count, bool *com_port_found, char *com_port_name)
{
	int ret;

	PL_DBG("Begin; product_id = '%s'; board_identification = %p.", product_id, board_identification);

	unsigned int i = 0;

	while (1)
	{
		PL_DBG("i = %u", i);

		ret = Enum_boards_(product_id, board_identification, count, com_port_found, com_port_name);

		PL_DBG("ret = %d", ret);

		if (ret != 2)
			break;

		i++;

		if (i >= 1000)
		{
			PL_DBG("ERROR: Exceeded the number of attempts to enumerate boards!");
			break;
		}

		usleep(1000);
	}

	PL_DBG("End; Return: %d.", ret);
	return ret;
}

int Read_from_com_port(int fd, char *buf, size_t count)
{
	PL_DBG("Begin; fd = %d; buf = %p; count = %ld.", fd, buf, count);

	ssize_t ret_ln;
	size_t num_bytes = 0;
	size_t n_bytes = count;

	while (n_bytes)
	{
		PL_DBG("num_bytes = %lu; n_bytes = %lu", num_bytes, n_bytes);

		ret_ln = read(fd, buf + num_bytes, n_bytes);
		if ((ret_ln < 0) || (ret_ln > n_bytes))
		{
			PL_DBG("ERROR %d (%s): Failed read from COM-port (ret_ln = %ld)!", errno, strerror(errno), ret_ln);
			PL_DBG("End (With ERROR!); Return: 1!");
			return 1;
		}

		PL_DBG("ret_ln = %ld", ret_ln);

		num_bytes += ret_ln;
		n_bytes -= ret_ln;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int Send_command_to_stm(const char *com_port_name, unsigned char command, __u32 *response)
{
	PL_DBG("Begin; com_port_name = '%s'; command = 0x%2.2X = %u.", com_port_name, command, command);

	int hCom = open(com_port_name, O_RDWR | O_SYNC);
	if (hCom == -1)
	{
		PL_DBG("ERROR %d (%s): Failed open COM-port!", errno, strerror(errno));
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	struct termios tty_prev, tty;

	int ret = tcgetattr(hCom, &tty);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed get COM-port attributes (ret = %d)!", errno, strerror(errno), ret);
		close(hCom);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	memcpy(&tty_prev, &tty, sizeof(tty));

	cfsetospeed(&tty, B115200);
	cfsetispeed(&tty, B115200);

	tty.c_cflag &= ~CSIZE;
	tty.c_cflag |= CS8;

	tty.c_cflag &= ~(CSTOPB | PARENB);
	tty.c_cflag |= (CREAD | CLOCAL);

	tty.c_lflag &= ~(ISIG | ICANON | ECHO | ECHOE);

	tty.c_iflag &= ~(IGNCR | ICRNL | IXON | IXANY | IXOFF);

	tty.c_oflag &= ~OPOST;

	tty.c_cc[VMIN] = 0x1;

	ret = tcsetattr(hCom, TCSANOW, &tty);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed set COM-port attributes (ret = %d)!", errno, strerror(errno), ret);
		close(hCom);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	unsigned char buf[4];
	ssize_t ret_ln;

	buf[0] = command;

	PL_DBG("Sending command...");

	ret_ln = write(hCom, buf, 1);
	if (ret_ln != 1)
	{
		PL_DBG("ERROR %d (%s): Failed send command to COM-port (ret_ln = %ld)!", errno, strerror(errno), ret_ln);
		tcsetattr(hCom, TCSANOW, &tty_prev);
		close(hCom);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	unsigned int response_size = 1;

	if (command == STM_GET_ID)
		response_size = 4;

	PL_DBG("Getting response...");

	size_t num_bytes = 0;
	size_t n_bytes = response_size;

	while (n_bytes)
	{
		PL_DBG("num_bytes = %lu; n_bytes = %lu", num_bytes, n_bytes);

		ret_ln = read(hCom, buf + num_bytes, n_bytes);
		if ((ret_ln < 0) || (ret_ln > n_bytes))
		{
			PL_DBG("ERROR %d (%s): Failed get response from COM-port (ret_ln = %ld)!", errno, strerror(errno), ret_ln);
			tcsetattr(hCom, TCSANOW, &tty_prev);
			close(hCom);
			PL_DBG("End (With ERROR!); Return: 1!");
			return 1;
		}

		PL_DBG("ret_ln = %ld", ret_ln);

		num_bytes += ret_ln;
		n_bytes -= ret_ln;
	}

	PL_DBG("Response: 0x%2.2X 0x%2.2X 0x%2.2X 0x%2.2X", buf[0], buf[1], buf[2], buf[3]);

	ret = tcsetattr(hCom, TCSANOW, &tty_prev);
	if (ret)
	{
		PL_DBG("WARNING %d (%s): Failed restore COM-port attributes (ret = %d)!", errno, strerror(errno), ret);
	}

	ret = close(hCom);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed close COM-port (ret = %d)!", errno, strerror(errno), ret);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (command == STM_GET_ID)
	{
		assert(response);

		*response = *((__u32 *)buf);
	}
	else
		if (buf[0] != STM_RESPONSE)
		{
			PL_DBG("ERROR: Wrong response to the command!");
			PL_DBG("End (With ERROR!); Return: 1!");
			return 1;
		}

	PL_DBG("End; Return: 0.");
	return 0;
}

int Get_nm_id(const char *nm_com_port_name, __u32 *nm_id)
{
	PL_DBG("Begin; nm_com_port_name = '%s'.", nm_com_port_name);

	PL_Board *tmpBoard;

	try
	{
		tmpBoard = new PL_Board(nm_com_port_name);
	}
	catch (...)
	{
		PL_DBG("ERROR: Failed create temporary board descriptor!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (tmpBoard->Get_board_id())
	{
		PL_DBG("ERROR: Failed get NM ID!");
		delete tmpBoard;
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	*nm_id = tmpBoard->Board_ID;

	delete tmpBoard;

	PL_DBG("End; Return: 0.");
	return 0;
}

int Check_board(PL_Board *board)
{
	PL_DBG("Begin.");

	if (!board)
	{
		PL_DBG("ERROR: Invalid 'board' (NULL)!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (board->Magic_code != BOARD_MAGIC_CODE_VAL)
	{
		PL_DBG("ERROR: Wrong board magic code 0x%8.8X (expect 0x%8.8X)!", board->Magic_code, BOARD_MAGIC_CODE_VAL);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int Check_and_capture_board(PL_Board *board)
{
	PL_DBG("Begin.");

	LOCK_LL

	if (Check_board(board))
	{
		PL_DBG("ERROR: Invalid 'board'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	board->Count_working_functions++;

	PL_DBG("Count of working functions: %u.", board->Count_working_functions);

	UNLOCK_LL

	PL_DBG("End; Return: 0.");
	return 0;
}

void Release_board(PL_Board *board)
{
	PL_DBG("Begin.");

	LOCK_LL

	assert(board->Count_working_functions);

	board->Count_working_functions--;

	PL_DBG("Count of working functions: %u.", board->Count_working_functions);

	UNLOCK_LL

	PL_DBG("End.");
	return;
}

int Check_access(PL_Access *access)
{
	PL_DBG("Begin.");

	if (!access)
	{
		PL_DBG("ERROR: Invalid 'access' (NULL)!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (access->Magic_code != PROCESSOR_MAGIC_CODE_VAL)
	{
		PL_DBG("ERROR: Wrong processor magic code 0x%8.8X (expect 0x%8.8X)!", access->Magic_code, PROCESSOR_MAGIC_CODE_VAL);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int Check_and_capture_access(PL_Access *access)
{
	PL_DBG("Begin.");

	LOCK_LL

	if (Check_access(access))
	{
		PL_DBG("ERROR: Invalid 'access'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	access->Count_working_functions++;

	PL_DBG("Count of working functions: %u.", access->Count_working_functions);

	UNLOCK_LL

	PL_DBG("End; Return: 0.");
	return 0;
}

void Release_access(PL_Access *access)
{
	PL_DBG("Begin.");

	LOCK_LL

	access->Count_working_functions--;

	PL_DBG("Count of working functions: %u.", access->Count_working_functions);

	UNLOCK_LL

	PL_DBG("End.");
	return;
}

int Check_and_capture_access_for_sync(PL_Access *access)
{
	PL_DBG("Begin.");

	LOCK_LL

	if (Check_access(access))
	{
		PL_DBG("ERROR: Invalid 'access'!");
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (access->Sync_in_progress)
	{
		PL_DBG("ERROR: Sync already in progress (Sync_in_progress = %d)!", access->Sync_in_progress);
		UNLOCK_LL
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	access->Count_working_functions++;
	access->Sync_in_progress = 1;

	PL_DBG("Count of working functions: %u.", access->Count_working_functions);

	UNLOCK_LL

	PL_DBG("End; Return: 0.");
	return 0;
}

void Release_access_for_sync(PL_Access *access)
{
	PL_DBG("Begin.");

	LOCK_LL

	access->Count_working_functions--;
	access->Sync_in_progress = 0;
	access->Sync_cancel = 0;

	PL_DBG("Count of working functions: %u.", access->Count_working_functions);

	UNLOCK_LL

	PL_DBG("End.");
	return;
}

int Check_mem_ranges(unsigned int procNo, PL_Addr addr, PL_Word len, int RW)
{
	PL_DBG("Begin; procNo = %u; addr = 0x%8.8X; len = 0x%8.8X; RW = %d.", procNo, addr, len, RW);

	if (!len)
	{
		PL_DBG("ERROR: len = 0!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (!(
		((procNo == 0) && ((addr >= C0_IMU_BEGIN) && (addr < C0_IMU_END) && (len <= (C0_IMU_END - addr)))) ||
		((procNo == 1) && ((addr >= C1_IMU_BEGIN) && (addr < C1_IMU_END) && (len <= (C1_IMU_END - addr)))) ||
		((addr >= C0_IMU_DUB_BEGIN) && (addr < C0_IMU_DUB_END) && (len <= (C0_IMU_DUB_END - addr))) ||
		((addr >= C1_IMU_DUB_BEGIN) && (addr < C1_IMU_DUB_END) && (len <= (C1_IMU_DUB_END - addr))) ||
		((addr >= SHMEM_BEGIN) && (addr < SHMEM_END) && (len <= (SHMEM_END - addr))) ||
		((addr >= USB_BEGIN) && (addr < USB_END) && (len <= (USB_END - addr))) ||
		((addr >= MDMACapb_BEGIN) && (addr < MDMACapb_END) && (len <= (MDMACapb_END - addr))) ||
		((addr >= EMIC_BEGIN) && (addr < EMIC_END) && (len <= (EMIC_END - addr))) ||
		((addr >= PER_BEGIN) && (addr < PER_END) && (len <= (PER_END - addr))) ||
		((addr >= EMI_BEGIN) && (addr < EMI_END) && (len <= (EMI_END - addr))) ))
	{
		PL_DBG("ERROR: Wrong memory range (addr = 0x%8.8X; len = 0x%8.8X)!", addr, len);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if ((RW) && (
		((procNo == 0) && ((addr >= C0_IMU_BEGIN) && (addr < C0_IMU_BEGIN_USER))) ||
		((procNo == 1) && ((addr >= C1_IMU_BEGIN) && (addr < C1_IMU_BEGIN_USER))) ||
		((addr >= C0_IMU_DUB_BEGIN) && (addr < C0_IMU_DUB_BEGIN_USER)) ||
		((addr >= C1_IMU_DUB_BEGIN) && (addr < C1_IMU_DUB_BEGIN_USER)) ||
		((addr >= SHMEM_BEGIN) && (addr < SHMEM_BEGIN_USER)) ))
	{
		PL_DBG("ERROR: Memory range (addr = 0x%8.8X; len = 0x%8.8X) intersect with RO area!", addr, len);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}
