// Load and exchange library for MC121.01
// Internal classes

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>
#include <assert.h>
#include <errno.h>

#include "libload_def.h"
#include "libload_int.h"

PL_Board::PL_Board(const char *pcCommPort):
	Count_opened_PL_Access(0),
	Com_port_in_use(0),
	Count_working_functions(0)
{
	int ret;

	PL_DBG("Begin; pcCommPort: '%s'.", pcCommPort);

	if (strlen(INDIVIDUAL_BOARD_LOCK_FILE_PREFIX) + strlen(pcCommPort) > PATH_MAX)
	{
		PL_DBG("ERROR: Too long COM-port file name!");
		throw 1;
	}

	char Individual_lock_file_name[PATH_MAX + 1];

	strcpy(Individual_lock_file_name, INDIVIDUAL_BOARD_LOCK_FILE_PREFIX);
	strcat(Individual_lock_file_name, pcCommPort);

	for (unsigned int i = strlen(INDIVIDUAL_BOARD_LOCK_FILE_PREFIX); i < strlen(Individual_lock_file_name); i++)
		if (Individual_lock_file_name[i] == '/')
			Individual_lock_file_name[i] = '_';

	PL_DBG("Individual_lock_file_name: '%s'", Individual_lock_file_name);

	hIndividual_lock_file = open(Individual_lock_file_name, O_CREAT, S_IRUSR | S_IRGRP | S_IROTH);
	if (hIndividual_lock_file == -1)
	{
		PL_DBG("ERROR %d (%s): Failed create or open individual board lock file!", errno, strerror(errno));
		throw 1;
	}

	PL_DBG("Capturing (shared) individual board lock...");

	ret = flock(hIndividual_lock_file, LOCK_SH);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed capture (shared) individual board lock (ret = %d)!", errno, strerror(errno), ret);
		close(hIndividual_lock_file);
		throw 1;
	}

	int hCom = open(pcCommPort, O_RDWR | O_SYNC);
	if (hCom == -1)
	{
		PL_DBG("ERROR %d (%s): Failed open COM-port!", errno, strerror(errno));
		close(hIndividual_lock_file);
		throw 1;
	}

	struct termios tty;

	ret = tcgetattr(hCom, &tty);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed get COM-port attributes (ret = %d)!", errno, strerror(errno), ret);
		close(hCom);
		close(hIndividual_lock_file);
		throw 1;
	}

	cfsetospeed(&tty, B115200);
	cfsetispeed(&tty, B115200);

	tty.c_cflag &= ~CSIZE;
	tty.c_cflag |= CS8;

	tty.c_cflag &= ~(CSTOPB | PARENB);
	tty.c_cflag |= (CREAD | CLOCAL);

	tty.c_lflag &= ~(ISIG | ICANON | ECHO | ECHOE);

	tty.c_iflag &= ~(IGNCR | ICRNL | IXON | IXANY | IXOFF);

	tty.c_oflag &= ~OPOST;

	tty.c_cc[VMIN] = 0x1;

	ret = tcsetattr(hCom, TCSANOW, &tty);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed set COM-port attributes (ret = %d)!", errno, strerror(errno), ret);
		close(hCom);
		close(hIndividual_lock_file);
		throw 1;
	}

	hCom_port = hCom;

	Magic_code = BOARD_MAGIC_CODE_VAL;

	PL_DBG("End.");
}

PL_Board::~PL_Board()
{
	int ret;

	PL_DBG("Begin.");

	Magic_code = 0x0;

	if (hCom_port != -1)
	{
		PL_DBG("Closing COM-port...");

		ret = close(hCom_port);
		if (ret)
		{
			PL_DBG("ERROR %d (%s): Failed close COM-port (ret = %d)!", errno, strerror(errno), ret);
		}
	}

	ret = close(hIndividual_lock_file);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed close individual board lock file!", errno, strerror(errno));
	}

	PL_DBG("End.");
}

int PL_Board::Get_board_id()
{
	PL_DBG("Begin.");

	if (Capture_port())
	{
		PL_DBG("ERROR: Failed capture COM-port!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (Send_command(READ_MEM, 0, BOARD_SERIAL_NUMBER_ADDR, 1))
	{
		PL_DBG("ERROR: Failed send command (for reading serial number)!");
		Release_port();
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (Read_from_com_port(hCom_port, (char *)(&Board_ID), 4))
	{
		PL_DBG("ERROR: Failed read board ID from COM-port!");
		Release_port();
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (Release_port())
	{
		PL_DBG("ERROR: Failed release COM-port!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int PL_Board::Switch_to_stm_if_no_more_used_and_close_com_port()
{
	int fl_ret, ret;

	PL_DBG("Begin.");

	fl_ret = flock(hIndividual_lock_file, LOCK_UN);
	if (fl_ret)
	{
		PL_DBG("ERROR %d (%s): Failed release individual board lock (fl_ret = %d)!", errno, strerror(errno), fl_ret);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("Try to capturing (exclusive and not blocking) individual board lock...");

	fl_ret = flock(hIndividual_lock_file, LOCK_EX | LOCK_NB);
	if (fl_ret && errno != EWOULDBLOCK)
	{
		PL_DBG("ERROR %d (%s): Failed capture individual board lock (fl_ret = %d)!", errno, strerror(errno), fl_ret);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("fl_ret = %d; errno = %d", fl_ret, errno);

	if (!fl_ret)
	{
		PL_DBG("Sending command to NM power off...");

		if (Send_command(NM_PWR_OFF, 0x0, 0x0, 0x0))
		{
			PL_DBG("ERROR: Failed send command (NM power off)!");
			PL_DBG("End (With ERROR!); Return: 1!");
			return 1;
		}
	}

	ret = close(hCom_port);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed close COM-port (ret = %d)!", errno, strerror(errno), ret);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	hCom_port = -1;

	if (!fl_ret)
	{
		PL_DBG("Waiting STM from NM...");

		bool stm_com_port_found = false;

		for (unsigned int i = 0; i < 100; i++)
		{
			PL_DBG("i = %u", i);

			usleep(100000);

			PL_board_identification bi = {PL_IDENTIFICATION_BY_BOARD_ID, NULL, Board_ID};

			if (Enum_boards(STM_PRODUCT_ID, &bi, NULL, &stm_com_port_found, NULL))
			{
				PL_DBG("ERROR: Failed enumerate STM boards after switching!");
				PL_DBG("End (With ERROR!); Return: 1!");
				return 1;
			}

			if (stm_com_port_found)
			{
				PL_DBG("STM from NM found.");

				break;
			}
		} // for i

		if (!stm_com_port_found)
		{
			PL_DBG("ERROR: STM from NM not found!");
			PL_DBG("End (With ERROR!); Return: 1!");
			return 1;
		}
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int PL_Board::Capture_port()
{
	PL_DBG("Begin.");

	while (__sync_fetch_and_or(&Com_port_in_use, 1))
		usleep(1);

	assert(Com_port_in_use == 1);

	PL_DBG("Capturing COM-port lock...");

	int ret = flock(hCom_port, LOCK_EX);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed capture COM-port lock (ret = %d)!", errno, strerror(errno), ret);
		__sync_fetch_and_and(&Com_port_in_use, 0);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int PL_Board::Release_port()
{
	PL_DBG("Begin.");

	assert(Com_port_in_use == 1);

	int ret = flock(hCom_port, LOCK_UN);
	if (ret)
	{
		PL_DBG("ERROR %d (%s): Failed release COM-port lock (ret = %d)!", errno, strerror(errno), ret);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	__sync_fetch_and_and(&Com_port_in_use, 0);

	PL_DBG("End; Return: 0.");
	return 0;
}

int PL_Board::Send_command(__u32 cmd, __u32 param1, __u32 param2, __u32 param3)
{
	PL_DBG("Begin; cmd = 0x%8.8X = %u; param1 = 0x%8.8X = %u; param2 = 0x%8.8X = %u; param3 = 0x%8.8X = %u.",
	                         cmd, cmd,        param1, param1,        param2, param2,        param3, param3);

	__u32 command[4];

	command[0] = cmd;
	command[1] = param1;
	command[2] = param2;
	command[3] = param3;

	ssize_t ret_ln = write(hCom_port, command, 16);
	if (ret_ln != 16)
	{
		PL_DBG("ERROR %d (%s): Failed write command to COM-port (ret_ln = %ld)!", errno, strerror(errno), ret_ln);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

#ifdef USE_SYNC_AFTER_WRITE
	sync();
#endif

	PL_DBG("End; Return: 0.");
	return 0;
}

PL_Access::PL_Access(unsigned int index, PL_Board *brd):
	board(brd),
	processorNo(index),
	Count_working_functions(0),
	Sync_cancel(0),
	Sync_in_progress(0)
{
	PL_DBG("Begin.");

	board->Count_opened_PL_Access++;

	PL_DBG("board->Count_opened_PL_Access = %u", board->Count_opened_PL_Access);

	Magic_code = PROCESSOR_MAGIC_CODE_VAL;

	PL_DBG("End.");
}

PL_Access::~PL_Access()
{
	PL_DBG("Begin.");

	Magic_code = 0x0;

	board->Count_opened_PL_Access--;

	PL_DBG("board->Count_opened_PL_Access = %u", board->Count_opened_PL_Access);

	PL_DBG("End.");
}

int PL_Access::Read(PL_Word *buf, PL_Addr addr, __u32 size)
{
	PL_DBG("Begin; addr = 0x%8.8X; size = 0x%8.8X.", addr, size);

	assert(size <= 0x3FFFFFFF);

	if (board->Send_command(READ_MEM, processorNo, addr, size))
	{
		PL_DBG("ERROR: Failed send command (read)!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	if (Read_from_com_port(board->hCom_port, (char *)buf, size << 2))
	{
		PL_DBG("ERROR: Failed read from COM-port!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	PL_DBG("End; Return: 0.");
	return 0;
}

int PL_Access::Write(PL_Word *buf, PL_Addr addr, __u32 size)
{
	PL_DBG("Begin; addr = 0x%8.8X; size = 0x%8.8X.", addr, size);

	assert(size <= 0x3FFFFFFF);

	if (board->Send_command(WRITE_MEM, processorNo, addr, size))
	{
		PL_DBG("ERROR: Failed send command (write)!");
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

	ssize_t ret_ln = write(board->hCom_port, buf, size << 2);
	if (ret_ln != (size << 2))
	{
		PL_DBG("ERROR %d (%s): Failed write to COM-port (ret_ln = %ld)!", errno, strerror(errno), ret_ln);
		PL_DBG("End (With ERROR!); Return: 1!");
		return 1;
	}

#ifdef USE_SYNC_AFTER_WRITE
	sync();
#endif

	PL_DBG("End; Return: 0.");
	return 0;
}

int PL_Access::Write1(PL_Word addr, PL_Word val)
{
	return Write(&val, addr, 1);
}
