// Loader for MC121.06

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include "nmrun.h"

using namespace std;

PL_board_identification Board_identification = {PL_ARBITRARY_INSTANCE};

unsigned int Core_for_features_number = 0;

PL_Board *Board;
PL_Access *Access[CORES_COUNT];

int f_Dump_start = 0;
int f_Dump_finish = 0;
int f_Run_program = 1;
int f_Run_program_special = 0;
int f_Return_NM_ret_val = 0;
int Quiet = 0; // Suppress output (except errors)

char *ABS_file_name[CORES_COUNT];

unsigned long Timeout_value = 0; // (mS)

PL_Addr Dump_memory_begin = Dump_memory_begin_default;
PL_Addr Dump_memory_size = Dump_memory_size_default;
unsigned int Dump_field_bits = 32; // Size of output words in bits (must be power of 2 and not greater than 64)

unsigned char recv_flags = 0x0; // (disp, sect, size, addr)
char *recv_file_name = NULL;
PL_Addr recv_addr = Dump_memory_begin_default;
unsigned long recv_size = Dump_memory_size_default;
char *recv_sect_name;
PL_Addr recv_disp = 0x0;

unsigned char send_flags = 0x0; // (disp, sect, size, addr)
char *send_file_name = NULL;
PL_Addr send_addr = Dump_memory_begin_default;
unsigned long send_size = 0; // 0 - Read entire file
char *send_sect_name;
PL_Addr send_disp = 0x0;

int f_Sync = 0;
int Sync_send, Sync_recv;

int main(int argc, char **argv)
{
	for (unsigned int i = 0; i < CORES_COUNT; i++)
		ABS_file_name[i] = NULL;

	Display_version(argc, argv);
	Parse_command_line(argc, argv);

	try
	{
		unsigned int Boards_count, Firmware_version_major, Firmware_version_minor;

		if (PL_GetBoardCount(&Boards_count) != PL_OK)
		{
			cout << "ERROR: Failed get count of boards!\n";
			exit(1);
		}

		if (!Boards_count)
		{
			cout << "ERROR: No boards found!\n";
			exit(1);
		}

		if (PL_GetBoardDescEx(&Board_identification, &Board) != PL_OK)
		{
			cout << "ERROR: Failed open board!\n";
			exit(1);
		}

		if (!Quiet)
			if (PL_GetFirmwareVersion(Board, &Firmware_version_major, &Firmware_version_minor) == PL_OK)
				cout << "Firmware v" << Firmware_version_major << "." << Firmware_version_minor << "\n";
			else
				cout << "WARNING: Failed get firmware version!\n";

		for (unsigned int i = 0; i < CORES_COUNT; i++)
			if (PL_GetAccess(Board, i, &Access[i]) != PL_OK)
			{
				cout << "ERROR: Failed access core " << i << "!\n";
				for (unsigned int j = 0; j < i; j++)
					PL_CloseAccess(Access[j]);
				PL_CloseBoardDesc(Board);
				exit(1);
			}

		if (f_Dump_start)
		{
			if (Dump_shared_memory(".dms"))
				cout << "ERROR: Failed create dump of shared memory!\n";
		}
		
		if (send_file_name)
		{
			if (Send_binary_file_to_board())
			{
				cout << "ERROR: Failed copy binary file to board!\n";
				for (unsigned int acc = 0; acc < CORES_COUNT; acc++)
					PL_CloseAccess(Access[acc]);
				PL_CloseBoardDesc(Board);
				exit(1);
			}
		}

		PL_Word Status[CORES_COUNT], ABS_return_value[CORES_COUNT];

		if (f_Run_program)
		{
			int ret;

			for (unsigned int i = 0; i < CORES_COUNT; i++)
				if (ABS_file_name[i])
				{
					if (!f_Run_program_special)
						ret = PL_LoadProgramFile(Access[i], ABS_file_name[i]);
					else
						ret = PL_LoadProgramFile_Int(Access[i], ABS_file_name[i]);

					if (ret != PL_OK)
					{
						if (ret == PL_FILE)
							cout << ABS_file_name[i] << " :: ERROR: Failed find user program file!\n";
						else if (ret == PL_BADADDRESS)
							cout << ABS_file_name[i] << " :: ERROR: Invalid user program configuration!\n";
						else
							cout << ABS_file_name[i] << " :: ERROR: Failed load program into core " << i << "!\n";
						for (unsigned int acc = 0; acc < CORES_COUNT; acc++)
							PL_CloseAccess(Access[acc]);
						PL_CloseBoardDesc(Board);
						exit(1);
					}

					if (!Quiet)
					{
						cout << "Start user program on core " << i << "...\n";
						cout.flush();
					}
				}

			if (f_Run_program_special)
			{
				for (unsigned int acc = 0; acc < CORES_COUNT; acc++)
					PL_CloseAccess(Access[acc]);
				PL_CloseBoardDesc(Board);
				exit(0);
			}

			if (f_Sync)
			{
				if (!Quiet)
					cout << "Call barrier synchronization...\n" << flush;

				if (PL_Sync(Access[Core_for_features_number], Sync_send, &Sync_recv) == PL_OK)
				{
					cout << ABS_file_name[Core_for_features_number] << " :: Barrier synchronization value: ";
					cout << dec << Sync_recv;
					cout << " = 0x" << hex << uppercase << Sync_recv << ".\n";
				}
				else
					cout << "ERROR: Failed call barrier synchronization!\n";
			}

			struct timeval tv;
			unsigned int start_time;
			unsigned int current_time;

			if (Timeout_value)
			{
				gettimeofday(&tv, NULL);
				start_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
			}

			for (unsigned int i = 0; i < CORES_COUNT; i++)
				if (ABS_file_name[i])
				{
					// Waiting end program on core i
					while (1)
					{
						if (PL_GetStatus(Access[i], &Status[i]) != PL_OK)
						{
							cout << "ERROR: Failed get program status of core " << i << "!\n";
							for (unsigned int acc = 0; acc < CORES_COUNT; acc++)
								PL_CloseAccess(Access[acc]);
							PL_CloseBoardDesc(Board);
							exit(1);
						}

						if (Status[i] & PROGRAM_FINISHED)
							break;

						if (Timeout_value)
						{
							gettimeofday(&tv, NULL);
							current_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
							if ((current_time - start_time) > Timeout_value)
								break;
						}

						usleep(1000);
					}

					if (Status[i] & PROGRAM_FINISHED)
					{
						if (PL_GetResult(Access[i], &ABS_return_value[i]) != PL_OK)
						{
							cout << "ERROR: Failed get program return value of core " << i << "!\n";
							for (unsigned int acc = 0; acc < CORES_COUNT; acc++)
								PL_CloseAccess(Access[acc]);
							PL_CloseBoardDesc(Board);
							exit(1);
						}
					}
					else
						cout << ABS_file_name[i] << " :: ERROR: Timeout for program on core " << i << "!\n";
				}
		} // f_Run_program

		if (f_Dump_finish)
		{
			if (Dump_shared_memory(".dme"))
				cout << "ERROR: Failed create dump of shared memory!\n";
		}
		
		if (recv_file_name)
		{
			if (Receive_binary_dump_from_board())
				cout << "ERROR: Failed create binary dump of shared memory!\n";
		}

		for (unsigned int acc = 0; acc < CORES_COUNT; acc++)
			PL_CloseAccess(Access[acc]);
		PL_CloseBoardDesc(Board);

		if (f_Run_program && (!Quiet))
			for (unsigned int i = 0; i < CORES_COUNT; i++)
				if (ABS_file_name[i] && (Status[i] & PROGRAM_FINISHED))
				{
					cout << ABS_file_name[i] << " :: Core " << i << " return ";
					cout << dec << ABS_return_value[i];
					cout << " = 0x" << hex << uppercase << ABS_return_value[i] << ".\n";
				}

		if (f_Return_NM_ret_val)
			return ABS_return_value[Core_for_features_number];
	}
	catch (...)
	{
		cout << "ERROR: Unhandled exception!\n";
		exit(1);
	}

	return 0;
}
